import filecmp
import shutil
from pathlib import Path
from typing import Any, Tuple

from click import BaseCommand
from click.testing import CliRunner, Result


def click_run(cli: BaseCommand, *args: Any, **kwargs: Any) -> Result:
    runner = CliRunner()
    return runner.invoke(cli, *args, **kwargs)


def purge(*paths: Path) -> None:
    for p in paths:
        if p.is_dir():
            shutil.rmtree(p)
        elif p.is_file():
            p.unlink()


def cmp_outs(filename: str, paths: Tuple[Path, Path]) -> bool:
    file_one, file_two = (filepath / filename for filepath in paths)
    return filecmp.cmp(file_one, file_two)



from pathlib import Path
from typing import Dict, Tuple

import polars as pl

from .options import PycashierOpts
from .term import term


def gen_queries(sample: str, file: Path) -> pl.LazyFrame:
    return pl.scan_csv(file, separator="\t").with_columns(
        sample=pl.lit(sample),
        percent=(pl.col("count") / pl.col("count").sum() * 100).round(5),
    )


def parse_column_not_found_error(message: str) -> Tuple[str, str]:
    contents = message.splitlines()
    file = contents[4].split()[2]
    return contents[0], file


def receipt(files: Dict[str, Path], opts: PycashierOpts) -> None:
    term.log.info(f"Combining output files for {len(files)} samples.")
    term.log.debug("samples: " + ", ".join(files))

    lzdf = pl.concat(gen_queries(sample, file) for sample, file in files.items())
    if not opts.no_overlap:
        lzdf = lzdf.join(
            lzdf.group_by(pl.col("barcode"))
            .agg(
                pl.col("sample").alias("samples").sort(),
                pl.col("sample").len().alias("n_samples"),
            )
            .with_columns(pl.col("samples").list.join(";")),
            on=pl.col("barcode"),
        ).sort("sample", "count", "barcode", descending=True)
    try:
        lzdf.collect().write_csv(opts.output, separator="\t")
    except pl.ColumnNotFoundError as e:
        col, file = parse_column_not_found_error(e.args[0])
        term.log.error(f"missing column [b red]{col}[/] in [b]{file}[/]")
        term.log.error(f"check files in input directory: [b]{opts.input_}[/]")
        term.quit()
    except pl.ComputeError:
        term.log.error("unexpected error processing inputs")
        term.log.error(
            f"ensure [b]{opts.input_}[/] contains [b]pycashier extract[/] outputs"
        )
        term.quit()



import importlib.util
import sys
from pathlib import Path
from shutil import which
from typing import Dict, List, Optional

from rich import box
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from .term import term

PACKAGES = ["cutadapt", "fastp", "starcode", "pysam"]
CMD_PACKAGES: Dict[str, List[str]] = {
    "": PACKAGES,
    "receipt": [],
    "merge": ["fastp"],
    "extract": ["fastp", "cutadapt", "starcode"],
    "scrna": ["pysam", "cutadapt"],
}


def generate_table(command: str, pkg_map: Dict[str, Optional[str]]) -> Table:
    table = Table(box=box.SIMPLE, header_style="bold cyan", collapse_padding=True)
    table.add_column("package", justify="center")
    table.add_column("installed", justify="center")
    table.add_column("path")

    for name, exists in pkg_map.items():
        table.add_row(
            name,
            "[green]yes[/green]" if exists else "[red]no[/red]",
            Text(str(exists if exists else "Not found"), overflow="fold"),
            style="bold" if name in CMD_PACKAGES[command] else "dim",
        )
    return table


def check_file_permissions() -> None:
    try:
        # attempt to trigger PermissionError
        Path("pycashier.toml").is_file()
    except PermissionError:
        term.print(
            "[PycashierPermissionError] reading and writing from current directory not possible",
            err=True,
        )
        term.print(
            "If you are using docker please supply user flag, for example `-u $(id -u):$(id -g)`"
        )
        term.quit()


def pre_run_check(command: str = "", show: bool = False) -> None:
    """Check for runtime dependencies

    Args:
        command: Name of pycashier subcommand.
        show: If true, show table regardless.
    """
    pkg_locations = {name: find_tool(name) for name in PACKAGES}
    cmd_pkg_locations = {k: pkg_locations[k] for k in CMD_PACKAGES[command]}

    if None in cmd_pkg_locations.values() or show:
        table = generate_table(command, pkg_locations)

        term.print(
            Panel(
                table,
                title="Dependencies",
            ),
        )
        term.print(f"python exe: [bold]{sys.executable}[/bold]")

        term.print(
            "It's recommended to install pycashier within a conda environment.\n"
            "See the repo for details: [link]https://github.com/brocklab/pycashier[/link]",
        )
        if None in pkg_locations.values():
            term.print(
                f"\n[red bold] FAILED PRE-RUN CHECKS for [hl]pycashier {command}[/hl]!\n",
            )
            term.quit()

    check_file_permissions()


def find_tool(name: str) -> Optional[str]:
    """Check whether `name` is on PATH and marked as executable.

    Args:
        name: Executable/dependency.
    Returns:
        True for success, False otherwise
    """

    if not name == "pysam":
        return which(name)

    spec = importlib.util.find_spec("pysam")
    if spec:
        return spec.origin

    return None



from __future__ import annotations

import sys
from collections import Counter
from pathlib import Path
from typing import Any, Generator, List

import click

from .config import save_params
from .merge import get_pefastqs
from .options import PycashierOpts
from .receipt import receipt
from .sample import ExtractSample, MergeSample, Sample, ScrnaSample
from .term import term
from .termui import confirm_extract_samples, confirm_samples, print_params
from .utils import filter_input_by_sample


class Pycashier:
    def __init__(self, ctx: click.Context, save_config: bool, **kwargs: Any) -> None:
        self.opts = PycashierOpts(**kwargs)
        if not (parent := self.opts.pipeline.parent).is_dir():
            term.print(
                f"[InputError] pipeline parent directory, {parent}, does not exist.",
                err=True,
            )
            term.quit()
        self.opts.pipeline.mkdir(exist_ok=True)
        term.set_logger(self.opts.log_file, self.opts.verbose)

        # must be after logger is initialized
        term.log.debug("pycashier command line:\n  " + " ".join(sys.argv))
        self.mode = str(ctx.info_name)
        self.check_duplicates = self.mode != "merge"
        term.mode(cmd=self.mode)
        if save_config:
            save_params(ctx)
        print_params(ctx)

    def _get_input_files(
        self,
        exts: List[str],
    ) -> List[Path]:
        """determine input files
        Args:
            samples: List of allowed samples.
            exts: Acceptable file extensions.

        Returns:
            List of fastq/sam files (may be gzipped).
        """

        candidate_files = [
            f for f in self.opts.input_.iterdir() if not f.name.startswith(".")
        ]

        if not candidate_files:
            term.print(
                f"[InputError]: Source dir: {self.opts.input_}, appears to be empty...",
                err=True,
            )
            term.quit()

        bad_files = [
            f.name
            for f in candidate_files
            if not any(f.name.endswith(suffix) for suffix in exts)
        ]
        symlinks = [f for f in candidate_files if f.is_symlink() and not f.exists()]
        duplicates = []

        if self.check_duplicates:
            duplicates = [
                name
                for name, count in Counter(
                    [f.name.split(".")[0] for f in candidate_files]
                ).items()
                if count > 1
            ]

        if bad_files:
            term.print(
                f"[InputError]: There is a non [bold]{'/'.join(exts)}[/] file in the provided input directory: "
                + "; ".join((f"[red]{file}[/]") for file in bad_files),
                err=True,
            )
        if symlinks:
            term.print(
                f"[InputError]: There is a symlink in the provided input directory: {';'.join((link.name for link in symlinks))} \n"
                "If using docker: Ensure that symlinks are resolved within the mounted volume "
                "in '/data' or pycashier within docker won't find them.",
                err=True,
            )

        if duplicates:
            term.print(
                "[InputError]: There appears to be duplicates files in input directory for sample(s): "
                f"{';'.join((dup for dup in duplicates))} \n"
                "See [b]pycashier merge[/b] for overlapping PE reads.",
                err=True,
            )

        if bad_files or symlinks or duplicates:
            term.quit()

        if self.opts.samples:
            candidate_files = filter_input_by_sample(
                candidate_files, self.opts.samples.split(",")
            )

        return candidate_files

    def _log_samples(
        self, samples: List[ExtractSample] | List[MergeSample] | List[ScrnaSample]
    ) -> None:
        term.log.debug(
            f"processing {len(samples)} samples: "
            + ";".join((sample.name for sample in samples))
        )

    def _process_samples(
        self, samples: List[ExtractSample] | List[MergeSample] | List[ScrnaSample]
    ) -> Generator[Sample, None, None]:
        for sample in samples:
            term.log.debug(f"starting sample: {sample.name}")
            yield sample

    def _is_complete(
        self, samples: List[ExtractSample] | List[MergeSample] | List[ScrnaSample]
    ) -> None:
        if all((sample.completed for sample in samples)):
            term.quit()

    def _check_failure(
        self, samples: List[ExtractSample] | List[MergeSample] | List[ScrnaSample]
    ) -> None:
        failed = [sample for sample in samples if sample.status.value == 3]

        if failed:
            term.print(
                f"\n\nFailed to complete {len(failed)} samples.\n"
                f"See [hl]{self.opts.log_file}[/] for more info.",
                err=True,
            )

    def extract(
        self,
        ctx: click.Context,
        yes: bool,
        **kwargs: Any,
    ) -> None:
        """
        extract DNA barcodes from a directory of fastq files

        \b
        Sample names should be delimited with a ".", such as `[hl][yellow]<sample>[/yellow].raw.fastq[/]`,
        anything succeeding the first period will be ignored by `[hl]pycashier[/]`.

        If your data is paired-end with overlapping barcodes,
        see `[hl]pycashier merge[/]`.
        """

        # validate that filter count and filter percent aren't both defined
        self.opts.update_filter(ctx)

        with term.cash_in(f"checking {self.opts.pipeline}"):
            samples = [
                ExtractSample(fastq=f, opts=self.opts)
                for f in self._get_input_files(
                    exts=[".fastq", ".fastq.gz"],
                )
            ]

        confirm_extract_samples(samples, self.opts)
        self._is_complete(samples)

        self.opts.output.mkdir(exist_ok=True)

        samples = [sample for sample in samples if not sample.completed]
        self._log_samples(samples)
        for sample in self._process_samples(samples):
            sample.pipeline()
        self._check_failure(samples)

    def merge(
        self,
    ) -> None:
        """
        merge overlapping paired-end reads using fastp
        \n\n\n
        Simple wrapper over `[hl]fastp[/]` to combine R1 and R2 from PE fastq files.
        \n\n\n
        """

        samples = [
            MergeSample(fastqR1=fastqs["R1"], fastqR2=fastqs["R2"], opts=self.opts)
            for s, fastqs in get_pefastqs(
                self._get_input_files(
                    exts=[".fastq", ".fastq.gz"],
                )
            ).items()
        ]

        confirm_samples(samples, self.opts)

        self._is_complete(samples)
        self.opts.output.mkdir(exist_ok=True)

        samples = [sample for sample in samples if not sample.completed]
        for sample in samples:
            sample.pipeline()

        self._check_failure(samples)

    def scrna(
        self,
    ) -> None:
        """
        extract expressed DNA barcodes from scRNA-seq
        \n
        \b
        Designed for interoperability with 10X scRNA-seq workflow.
        After processing samples with `[hl]cellranger[/]` resulting
        bam files should be converted to sam files using `[hl]samtools[/]`.
        \n
        [i]NOTE[/]: You can speed this up by providing a sam file with only
        the unmapped reads.
        """
        samples = [
            ScrnaSample(sam=f, opts=self.opts)
            for f in self._get_input_files(exts=[".sam"])
        ]
        confirm_samples(samples, self.opts)
        self._is_complete(samples)
        self.opts.output.mkdir(exist_ok=True)

        samples = [sample for sample in samples if not sample.completed]
        for sample in samples:
            sample.pipeline()
        self._check_failure(samples)

    def receipt(
        self,
    ) -> None:
        """
        combine and summarize outputs of [hl]extract[/]
        """

        files = {f.name.split(".")[0]: f for f in self._get_input_files(exts=[".tsv"])}
        with term.cash_in("calculating"):
            receipt(files, self.opts)



project = "Pycashier"
copyright = "2024, BrockLab"
author = "BrockLab"

extensions = ["myst_parser", "sphinx_copybutton", "sphinx_tabs.tabs"]
myst_enable_extensions = ["colon_fence", "deflist"]
myst_heading_anchors = 1


templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


html_theme = "shibuya"
# html_static_path = ["_static"]
# TODO: make a logo?
# html_logo = "../assets/logo.svg"

html_theme_options = {
    "github_url": "https://github.com/brocklab/pycashier",
    "nav_links": [
        {"title": "Documentation", "url": "installation"},
        {"title": "CLI Reference", "url": "cli"},
    ],
}



import re
from pathlib import Path
from typing import Dict, List

from .term import term


def get_pefastqs(fastqs: List[Path]) -> Dict[str, Dict[str, Path]]:
    """parse input files for paired-end sequences

    Args:
        fastqs: list of fastq files from input directory
    Returns:
        dictionary of sample and paired-end fastq files
    """
    pefastqs: Dict[str, Dict[str, Path]] = {}
    for f in fastqs:
        m = re.search(r"(?P<sample>.+?)\..*(?P<read>R[1-2])\..*fastq(?:.gz)?", f.name)
        if m:
            sample, read = m.groups()
            pefastqs.setdefault(sample, {})

            if read not in pefastqs[sample]:
                pefastqs[sample][read] = f
            else:
                term.print(
                    f"[MergeError]: detected multiple [hl]{read}[/] files for [hl]{sample}[/]\n"
                    f"files: [b]{f}[/] and [b]{pefastqs[sample][read]}[/]",
                    err=True,
                )
                term.quit()
        else:
            term.print(
                f"[MergeError]: failed to obtain sample/read info from [b]{f}[/]\n",
                "Merge mode expects fastq(.gz) files with R1 or R2 in the name",
                err=True,
            )
            term.quit()

    for sample, reads in pefastqs.items():
        if len(reads) != 2:
            term.print(
                "[MergeError]: please ensure there is and R1 and R2 for all samples"
            )
            term.print("[MergeError]: detected the following samples")
            term.print(pefastqs)
            term.quit()

    return pefastqs



from __future__ import annotations

from pathlib import Path

import polars as pl

try:
    import pysam
except ImportError:
    pass

from rich.progress import Progress, SpinnerColumn, TimeElapsedColumn

from .term import term


# pysam types are finicky
def sam_to_name_labeled_fastq(
    sample: str, sam_file: Path, out_file: Path
) -> bool | None:
    """convert sam file to metadata labeled fastq

    Args:
        sample: Name of sample.
        sam_file: Sam file to convert.
        out_file: Converted fastq file.
        status: Rich.console status to suspend for stderr printing.
    """

    # if the file is a sam file this is the only way I can find in the
    # pysam API to get the total number of reads
    # we really only need this for the progess bar though
    try:
        with pysam.AlignmentFile(  # type: ignore
            str(sam_file.absolute()),
            "r",
            check_sq=False,
            check_header=False,
        ) as sam:
            with term.process("getting sam size"):
                sam_length = sam.count()
    except ValueError:
        term.log.error(f"Couldn't load sam file:{sam_file}. Is it the correct format?")
        return True

    # we don't care about indicies or genomes. since cutadapt will do the heavy lifting here
    with open(out_file, "w") as f_out, pysam.AlignmentFile(
        str(sam_file.absolute()),
        "r",
        check_sq=False,
        check_header=False,  # type: ignore
    ) as sam:
        with term._no_status(), Progress(
            SpinnerColumn("point", style="bright_magenta"),
            *Progress.get_default_columns(),
            "Elapsed:",
            TimeElapsedColumn(),
            transient=True,
            console=term._console,
        ) as progress:
            task = progress.add_task("sam -> fastq", total=sam_length)

            for record in sam.fetch(until_eof=True):
                tagdict = dict(record.tags)  # type: ignore
                cell_barcode = None
                if "CB" in tagdict.keys():
                    cell_barcode = tagdict["CB"].split("-")[0]
                elif "CR" in tagdict.keys():
                    cell_barcode = tagdict["CR"]

                umi = None
                if "UB" in tagdict.keys():
                    umi = tagdict["UB"]
                elif "UR" in tagdict.keys():
                    umi = tagdict["UR"]

                if cell_barcode and umi:
                    qualities = record.query_qualities
                    ascii_qualities = "".join([chr(q + 33) for q in qualities])  # type: ignore

                    f_out.write(f"@{record.query_name}_{umi}_{cell_barcode}\n")
                    f_out.write(f"{record.query_sequence}\n+\n{ascii_qualities}\n")

                progress.advance(task)


def labeled_fastq_to_tsv(in_file: Path, out_file: Path) -> bool | None:
    """convert labeled fastq to tsv

    Args:
        in_file: Fastq to convert to tsv.
        out_file: Converted tsv file.
    Return:
        1 if failure
    """
    try:
        (
            pl.scan_csv(in_file, has_header=False, separator="\t")
            .select(
                pl.all()
                .gather_every(4)
                .str.splitn("_", 3)
                .struct.rename_fields(["info", "umi", "cell"])
                .alias("info"),
                pl.all().gather_every(4, offset=1).alias("barcode"),
            )
            .unnest("info")
            .collect()
            .write_csv(out_file, separator="\t")
        )
    except pl.ComputeError:
        term.log.error(
            f"failed to convert fastq to tsv: {in_file}\n"
            "ensure fastq was not corrupted and contains all reads"
        )
        return True
    except pl.NoDataError:
        term.log.error(
            f"failed to convert fastq to tsv: {in_file}\n"
            "no reads found, check cutadapt output"
        )
        return True



import typing
from pathlib import Path

import pytest
import tomlkit
from pycashier.cli import extract
from utils import click_run, purge

TEST_DIR = Path(__file__).parent
OUTS_DIR, REF_DIR, MERGED_DIR, PIPELINE_DIR = (
    TEST_DIR / "data" / name
    for name in ("outs", "reference", "mergedfastqs", "pipeline")
)


@pytest.mark.parametrize("config_type", ["full", "explicit"])
def test_pycashier_extract_config(config_type: str) -> None:
    config_path = Path(TEST_DIR / "data" / "pycashier.toml")
    purge(OUTS_DIR, PIPELINE_DIR, config_path)
    config_path.touch(exist_ok=True)
    # shutil.copyfile(Path(TEST_DIR / "data" / "reference" / "test.toml"), config)

    result = click_run(
        extract,
        # must provide input...
        [
            "-i",
            str(REF_DIR / "rawfastqgzs"),
            "-e",
            0.15,
            "-fc",
            5,
            "-c",
            config_path,
            "--save-config",
            config_type,
        ],
        input="n\n",
    )
    assert result.exit_code == 0
    check_config("extract")


def read_config(p: Path) -> tomlkit.TOMLDocument:
    return tomlkit.loads(p.read_text())


# tomlkit has weird types
@typing.no_type_check
def check_config(section: str) -> None:
    ref = read_config(REF_DIR / "pycashier.toml")
    config = read_config(TEST_DIR / "data" / "pycashier.toml")
    for k, v in ref[section].items():
        if k in config[section]:
            assert config[section][k] == v



from __future__ import annotations

import shlex
import subprocess
from pathlib import Path
from typing import Dict, List

import click
import polars as pl

from .term import term


def filter_input_by_sample(
    candidate_files: List[Path], samples: List[str]
) -> List[Path]:
    """filter candidate fastqs based on user provided samples
    Args:
        candidate_fastqs: all fastq files found in input direcotry
        samples: user provided list of samples
    Returns:
        returns list of accepted fastqs
    """

    found, files, ignored = [], [], []
    for f in candidate_files:
        name, *_ = f.name.split(".")
        if name in samples:
            found.append(name)
            files.append(f)
        else:
            ignored.append(name)

    not_found = set(samples).difference(found)
    if not_found:
        term.print(f"[InputError]: Unknown sample(s) -> {not_found}", err=True)
        term.quit()

    if ignored:
        term.print(f"[dim]ignoring {len(set(ignored))} samples")

    return files


def fastq_to_tsv(in_file: Path, out_file: Path) -> bool | None:
    """convert fastq file to tsv

    Args:
        in_file: Fastq file to convert.
        out_file: TSV file to write to.
    """

    try:
        pl.scan_csv(in_file, has_header=False, separator="\t").select(
            pl.all().gather_every(4).alias("info"),
            pl.all().gather_every(4, offset=1).alias("barcode"),
        ).collect().write_csv(out_file, separator="\t")
    except pl.ComputeError:
        term.log.error(
            f"failed to convert fastq to tsv: {in_file}\n"
            "ensure fastq was not corrupted and contains all reads"
        )
        return True
    except pl.NoDataError:
        term.log.error(
            f"failed to convert fastq to tsv: {in_file}\n"
            "no reads found, check cutadapt output"
        )
        return True


def extract_csv_column(csv_file: Path, out_file: Path) -> None:
    """get column from csv file

    Args:
        csv_file: File to extract column from.
        out_file: File to write column to.
    """
    pl.scan_csv(csv_file, separator="\t").select(pl.col("barcode")).collect().write_csv(
        out_file, separator="\t", include_header=False
    )


def get_filter_count(file_in: Path, filter_percent: float, quit: bool = False) -> int:
    """calculate filter cutoff

    Args:
        file_in: Clustered lineage counts to filter.
        filter_percent: Percent cutoff of total reads count.
        quit: exit if failure to get filter cutoff
    Returns:
        Minimum nominal cutoff value.
    """
    try:
        sum = (
            pl.scan_csv(
                file_in,
                separator="\t",
                has_header=False,
                new_columns=("barcode", "count"),
            )
            .select(pl.col("count").sum())
            .collect(streaming=True)
            .item()
        )
    except pl.NoDataError:
        term.log.error(
            f"Failed to determine filter cutoff for empty file {file_in}.\n"
            " Please remove it and try again."
        )
        term.quit()
    return int(
        round(
            sum * filter_percent / 100,
            0,
        )
    )


def validate_filter_args(ctx: click.Context) -> Dict[str, float]:
    """validate filter argument from config and CLI

    Args:
        ctx: Click context.
    Returns:
        Dictionary defining the filter type and value.
    """
    if ctx.params["filter_count"] or ctx.params["filter_count"] == 0:
        if ctx.get_parameter_source("filter_percent").value == 3:  # type: ignore
            ctx.params["filter_percent"] = None
            del ctx.params["filter_percent"]
            return {"filter_count": ctx.params["filter_count"]}
        else:
            raise click.BadParameter(
                "`--filter-count` and `--filter-percent` are mutually exclusive"
            )
    else:
        del ctx.params["filter_count"]
        return {"filter_percent": ctx.params["filter_percent"]}


def exit_status(p: subprocess.CompletedProcess, file: Path) -> bool:
    """check command exit status and file size

    Args:
        p: Completed subprocess
        file: File to check for nonzero size.
    Returns:
        True for success, False otherwise.
    """
    return True if p.returncode != 0 or file.stat().st_size == 0 else False


def run_cmd(
    command: str,
    sample: str,
    output: Path,
    verbose: bool,
) -> bool | None:
    """run a subcommand

    Args:
        command: Subcommand to be run in subprocess.
        sample: Name of sample.
        output: file of immediate output.
        verbose: If true, print subcommand output.
    Returns:
        exit code
    """
    cmd_name = command.split()[0]
    p = subprocess.run(
        shlex.split(command),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )

    # remove 'progress: ##%' output from starcode
    stdout = (
        p.stdout
        if not cmd_name == "starcode"
        else "\n".join(
            [line for line in p.stdout.splitlines() if not line.startswith("progress")]
        )
    )

    term.log.debug("subcommand:\n  [b]" + command)
    term.log.debug(
        "subcommand output:\n"
        + "\n".join(("[b]| [/]" + line for line in stdout.splitlines()))
    )
    if exit_status(p, output):
        term.print(
            f"[{cmd_name.capitalize()}Error]: Subprocess for sample failed: [green]{sample}[/green]",
            err=True,
        )
        return True


def check_output(file: Path, message: str) -> bool:
    """check for output file and print message

    Args:
        file: Resulting file of step.
        message: Text to display related to step.
    Returns:
        True for success, False otherwise.
    """
    exists = file.is_file()

    if not exists:
        term.log.debug("missing output file: " + file.name)
        term.log.debug(message)
    else:
        term.log.debug("found output file: " + file.name)
    return exists



from __future__ import annotations

import logging
import shutil
import sys
from contextlib import contextmanager
from logging.handlers import RotatingFileHandler
from pathlib import Path
from textwrap import dedent
from typing import Any, Generator, Optional

from rich.console import Console
from rich.highlighter import NullHighlighter, RegexHighlighter
from rich.logging import RichHandler
from rich.panel import Panel
from rich.prompt import Confirm
from rich.status import Status
from rich.text import Text
from rich.theme import Theme

MAX_WIDTH = 110

theme = Theme(
    {
        "hl": "bold cyan",
        "line": "bright_magenta",
        "border": "red",
        # click-rich-help defaults
        "header": "bold italic cyan",
        "option": "bold yellow",
        "metavar": "green",
        "default": "dim",
        "required": "dim red",
    },
    inherit=True,
)


class StreamFormatter(logging.Formatter):
    """Rich-enabled logging formatter"""

    def __init__(self, verbose: bool) -> None:
        super().__init__()
        self.FORMATS = {
            **{
                level: (f"[{color}]%(levelname)-7s[/]" " %(message)s")
                for level, color in {
                    logging.DEBUG: "dim",
                    logging.WARNING: "yellow",
                    logging.ERROR: "red",
                    logging.CRITICAL: "red",  # not used
                }.items()
            },
            logging.INFO: "%(message)s",
        }
        if verbose:
            self.FORMATS.update({logging.INFO: "[text]%(levelname)-7s[/] %(message)s"})

    def format(self, record: logging.LogRecord) -> str:
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = MutlilineFormatter(log_fmt)
        return formatter.format(record)


class FileFormatter(logging.Formatter):
    """strip rich markdown from log messages"""

    def format(self, record: logging.LogRecord) -> str:
        formatter = MutlilineFormatter(self._fmt)
        record.msg = Text.from_ansi(Text.from_markup(record.msg).plain).plain
        return formatter.format(record)


class MutlilineFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        outlines = []
        lines = (save_msg := record.msg).splitlines()
        for line in lines:
            record.msg = line
            outlines.append(super().format(record))
        record.msg = save_msg
        record.message = (output := "\n".join(outlines))
        return output


class ErrorHighlighter(RegexHighlighter):
    """Apply style to anything that looks like an error."""

    highlights = [r"(?P<error>\[\w+Error\])"]


class Term:
    """rich-based ui"""

    _status_name: str | None = None
    _status: Status | None = None

    def __init__(self, width: Optional[int] = None) -> None:
        self._process_msg = "default msg"
        self._console = Console(highlight=False, theme=theme, width=width)
        self._err_console = Console(
            theme=Theme({"hl": "bold cyan", "error": "bold red"}, inherit=True),
            stderr=True,
            highlighter=ErrorHighlighter(),
            width=width,
        )

    # use property ?
    def set_logger(self, log_file: Path, verbose: bool) -> None:
        if not (parent := log_file.parent).is_dir():
            term.print(
                f"[CliError] [code]{parent}[/] directory for log file does not exist. "
                f"Please create [code]{parent}[/] first.",
                err=True,
            )
            sys.exit(1)

        logger = logging.getLogger("pycashier")
        logger.setLevel(logging.DEBUG)

        ch = RichHandler(
            console=self._console,
            show_time=False,
            highlighter=NullHighlighter(),
            markup=True,
            show_level=False,
            show_path=False,
        )
        ch.setLevel(logging.INFO if not verbose else logging.DEBUG)
        ch.setFormatter(StreamFormatter(verbose))

        fh = RotatingFileHandler(log_file, maxBytes=10 * 1024 * 1024, backupCount=5)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(FileFormatter("%(asctime)s | %(levelname)8s | %(message)s"))

        logger.addHandler(ch)
        logger.addHandler(fh)
        self.log = logging.getLogger("pycashier")

    @contextmanager
    def _no_status(self) -> Generator[None, None, None]:
        try:
            if status := getattr(self, "_status", None):
                status.stop()
            yield
        finally:
            if status:
                status.start()

    def print(self, *args: Any, err: bool = False, **kwargs: Any) -> None:
        console = self._err_console if err else self._console
        with self._no_status():
            console.print(*args, **kwargs)

    def quit(self, code: int = 1) -> None:
        if status := getattr(self, "_status", None):
            status.stop()
        self._err_console.print("Exiting.")
        sys.exit(code)

    @contextmanager
    def cash_in(self, name: str) -> Generator[Status, None, None]:
        try:
            self._status_name = name
            self._status = self._console.status(
                name, spinner="point", spinner_style="bright_magenta"
            )
            self._status.start()
            yield self._status
        finally:
            if self._status:
                self._status.stop()
            self._status_name = None
            self._status = None

    def mode(self, cmd: str) -> None:
        term.log.info(f"[bold]pycashier [green]{cmd}")

    def confirm(self, *args: Any, **kwargs: Any) -> bool:
        return Confirm.ask(*args, console=self._console, **kwargs)

    def style_title(self, title: str) -> Text:
        return Text.from_markup(f"[reset][hl]{title}[/hl][/reset]")

    def textbox(self, text: str, title: str = "", *args: Any, **kwargs: Any) -> None:
        self.print(
            Panel.fit(
                Text(dedent(text), *args, *kwargs),
                title=self.style_title(title),
                title_align="left",
            )
        )

    @contextmanager
    def process(self, msg: str = "") -> Generator[None, None, None]:
        name = getattr(self, "_status_name")
        status = getattr(self, "_status")
        if msg:
            msg = " [dim]" + msg
        try:
            status.update(name + msg)
            yield
        finally:
            status.update(name)


cols = shutil.get_terminal_size().columns
term = Term(width=MAX_WIDTH if cols > MAX_WIDTH else cols)



import shutil
from typing import Any, Callable, Dict, List

import click
from click_rich_help import StyledGroup
from rich.traceback import install

from ._checks import pre_run_check
from .options import Option, optmap
from .pycashier import Pycashier
from .term import theme

install(suppress=[click], show_locals=True)


def add_options(
    options: List[Callable],
) -> Callable[[click.decorators.FC], click.decorators.FC]:
    def _add_options(func: click.decorators.FC) -> click.decorators.FC:
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


_help_groups_ref = {
    k: set(v)
    for k, v in {
        "Input/Output Options": ["--input", *optmap.long_by_category("input/output")],
        "Trim (Cutadapt) Options": [
            "--minimum-length",
            *optmap.long_by_category("trim"),
        ],
        "Cluster (Starcode) Options": optmap.long_by_category("cluster"),
        "Quality (Fastp) Options": optmap.long_by_category("quality"),
        "Merge Options": optmap.long_by_category("merge"),
        "Filter Options": optmap.long_by_category("filter"),
        "Receipt Options": ["--no-overlap"],
        "General Options": [
            *optmap.long_by_category("general"),
            "--help",
        ],
        "hidden": ["--skip-init-check"],
    }.items()
}


def get_help_groups(
    options: List[Option],
    extra_groups: List[str] = [],
) -> Dict[str, List[str]]:
    flags = [option.long() for option in options] + ["--help"]
    groups = ["Input/Output Options", *extra_groups, "General Options"]
    help_groups: Dict[str, List[str]] = {group: [] for group in groups}
    for group in groups:
        for flag in flags:
            if flag in _help_groups_ref["hidden"]:
                continue
            if flag in _help_groups_ref[group]:
                help_groups[group].append(flag)

    return help_groups


CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],
    max_content_width=shutil.get_terminal_size(fallback=(110, 24))[0],
)


@click.group(
    cls=StyledGroup,
    theme=theme,
    context_settings=CONTEXT_SETTINGS,
)
@click.version_option(package_name="pycashier", prog_name="pycashier")
def cli() -> None:
    """Cash in on DNA Barcode Tags
    \n\n\n
    See `[hl]pycashier COMMAND -h[/]` for more information.
    """
    pass


@cli.command(
    option_groups=get_help_groups(
        optmap.subcmds["extract"],
        extra_groups=[
            "Quality (Fastp) Options",
            "Trim (Cutadapt) Options",
            "Cluster (Starcode) Options",
            "Filter Options",
        ],
    ),
    help=Pycashier.extract.__doc__,
)
@add_options([option.get_click_option() for option in optmap.subcmds["extract"]])
@click.pass_context
def extract(ctx: click.Context, save_config: bool, **kwargs: Any) -> None:
    pycashier = Pycashier(ctx, save_config, **kwargs)
    pycashier.extract(ctx, **kwargs)


@cli.command(
    option_groups=get_help_groups(
        optmap.subcmds["merge"], extra_groups=["Merge Options"]
    ),
    help=Pycashier.merge.__doc__,
)
@add_options([option.get_click_option() for option in optmap.subcmds["merge"]])
@click.pass_context
def merge(ctx: click.Context, save_config: bool, **kwargs: Any) -> None:
    pycashier = Pycashier(ctx, save_config, **kwargs)
    pycashier.merge()


@cli.command(
    option_groups=get_help_groups(
        optmap.subcmds["scrna"],
        extra_groups=[
            "Trim (Cutadapt) Options",
        ],
    ),
    help=Pycashier.scrna.__doc__,
)
@add_options([option.get_click_option() for option in optmap.subcmds["scrna"]])
@click.pass_context
def scrna(ctx: click.Context, save_config: bool, **kwargs: Any) -> None:
    pycashier = Pycashier(ctx, save_config, **kwargs)
    pycashier.scrna()


@cli.command(
    option_groups=get_help_groups(
        optmap.subcmds["receipt"], extra_groups=["Receipt Options"]
    ),
    help=Pycashier.receipt.__doc__,
)
@add_options([option.get_click_option() for option in optmap.subcmds["receipt"]])
@click.pass_context
def receipt(ctx: click.Context, save_config: bool, **kwargs: Any) -> None:
    pycashier = Pycashier(ctx, save_config, **kwargs)
    pycashier.receipt()


@cli.command(hidden=True, help="perform check for pycashier dependencies")
def checks() -> None:
    pre_run_check(show=True)


def main() -> None:
    cli(prog_name="pycashier")


if __name__ == "__main__":
    main()



import filecmp
from pathlib import Path
from typing import List, Tuple

import pytest
from click import BaseCommand
from pycashier.cli import checks, cli, extract, merge, receipt, scrna
from utils import click_run, purge

TEST_DIR = Path(__file__).parent
OUTS_DIR, REF_DIR, MERGED_DIR, PIPELINE_DIR = (
    TEST_DIR / "data" / name
    for name in ("outs", "reference", "mergedfastqs", "pipeline")
)


def cmp_outs(filename: str, paths: Tuple[Path, Path]) -> bool:
    file_one, file_two = (filepath / filename for filepath in paths)
    return filecmp.cmp(file_one, file_two)


def test_help() -> None:
    for cmd in cli, extract, merge, scrna, receipt:
        result = click_run(cmd, ["--help"])
        assert result.exit_code == 0


def test_pycashier_checks() -> None:
    result = click_run(
        checks,
    )
    assert result.exit_code == 0


@pytest.mark.parametrize(
    ("cmd", "input_dir", "pipe_dir", "ref_dir", "outs_dir", "file_name", "options"),
    (
        (
            extract,
            REF_DIR / "rawfastqgzs",
            PIPELINE_DIR / "pipe-extract",
            REF_DIR / "outs",
            OUTS_DIR,
            "test.q30.barcodes.r3d1.min0_off1.tsv",
            [],
        ),
        (
            extract,
            REF_DIR / "rawfastqgzs",
            PIPELINE_DIR / "pipe-extract-fc0",
            REF_DIR / "outs",
            OUTS_DIR,
            "test.q30.barcodes.r3d1.min0_off1.tsv",
            ["--filter-count", "0"],
        ),
        (
            scrna,
            REF_DIR / "sams",
            PIPELINE_DIR / "pipe-scrna",
            REF_DIR / "outs-scrna",
            OUTS_DIR,
            "test.umi_cell_labeled.barcode.tsv",
            [],
        ),
        (
            merge,
            REF_DIR / "unmergedfastqgzs",
            PIPELINE_DIR / "pipe-merge",
            REF_DIR / "mergedfastqs",
            MERGED_DIR,
            "test.merged.raw.fastq",
            [],
        ),
    ),
)
def test_pycashier(
    cmd: BaseCommand,
    input_dir: Path,
    pipe_dir: Path,
    ref_dir: Path,
    outs_dir: Path,
    file_name: str,
    options: List[str],
) -> None:
    PIPELINE_DIR.mkdir(exist_ok=True, parents=True)
    purge(outs_dir, pipe_dir)

    result = click_run(
        cmd,
        ["-i", input_dir, "-o", outs_dir, "-p", pipe_dir, "-y", *options],
    )

    print(result.output)
    assert result.exit_code == 0
    assert cmp_outs(file_name, (ref_dir, outs_dir))


def test_pycashier_receipt() -> None:
    PIPELINE_DIR.mkdir(exist_ok=True, parents=True)
    outfile = TEST_DIR / "data/combined.tsv"
    purge(outfile, PIPELINE_DIR / "pipe-receipt")

    result = click_run(
        receipt,
        ["-i", REF_DIR / "outs", "-p", PIPELINE_DIR / "pipe-receipt", "-o", outfile],
    )

    assert result.exit_code == 0
    assert cmp_outs("combined.tsv", (REF_DIR, TEST_DIR / "data"))



from __future__ import annotations

import shutil
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Tuple

from .filters import read_filter
from .options import PycashierOpts
from .scrna import labeled_fastq_to_tsv, sam_to_name_labeled_fastq
from .term import term
from .utils import (
    check_output,
    fastq_to_tsv,
    get_filter_count,
    run_cmd,
)


class SampleStatus(Enum):
    COMPLETE = 0
    INCOMPLETE = 1
    WARN = 2
    FAIL = 3


def status_check(func: Callable) -> Callable:
    def wrapper(self: Sample, *args: Any, **kwargs: Any) -> None:
        if func(self, *args, **kwargs):
            self.status = SampleStatus.FAIL

    return wrapper


class Sample:
    def __init__(self, name: str, opts: PycashierOpts) -> None:
        self.opts = opts
        self.name = name
        self.status = (
            SampleStatus.COMPLETE
            if all(self.check().values())
            else SampleStatus.INCOMPLETE
        )
        self.steps: Tuple[Callable, ...]
        self.completed = self.status is SampleStatus.COMPLETE

    def check(self) -> Dict[str, bool]:
        raise NotImplementedError

    def finished(self, success: bool = True) -> None:
        symbol = (
            "[green]✔[/]"
            if self.status == SampleStatus.COMPLETE
            else ("[yellow]⚠[/]" if self.status == SampleStatus.WARN else "[red]✘[/]")
        )
        term.print(f"[b]{symbol} {self.name}")

    def pipeline(self) -> None:
        with term.cash_in(self.name):
            for step in self.steps:
                step()
                if self.status != SampleStatus.INCOMPLETE:
                    break
        if self.status == SampleStatus.INCOMPLETE:
            self.status = SampleStatus.COMPLETE
        self.finished()


class ExtractFiles:
    def __init__(self, name: str, opts: PycashierOpts) -> None:
        self.quality = opts.pipeline / f"{name}.q{opts.quality}.fastq"
        self.barcode_fastq = self.quality.with_suffix(".barcode.fastq")
        self.barcodes = self.quality.with_suffix(".barcodes.tsv")
        self.clustered = self.barcodes.with_suffix(
            f".r{opts.ratio}d{opts.distance}.tsv"
        )

    def final(self, opts: PycashierOpts) -> Optional[Path]:
        if not self.clustered.is_file():
            return None
        if opts.filter_percent:
            min_count = get_filter_count(self.clustered, float(opts.filter_percent))
        else:
            min_count = opts.filter_count
        return (opts.output / self.clustered.name).with_suffix(
            f".min{min_count}_off{opts.offset}.tsv"
        )


class ExtractSample(Sample):
    def __init__(self, fastq: Path, opts: PycashierOpts) -> None:
        name = fastq.name.split(".")[0]
        self.fastq = fastq
        self.files = ExtractFiles(name=name, opts=opts)
        self.steps = (
            self._filter,
            self._cutadapt,
            self._fast2tsv,
            self._starcode,
            self._read_filter,
        )
        super().__init__(name, opts)

    def check(self) -> Dict[str, bool]:
        exists = {}
        for name in ("quality", "barcodes", "clustered"):
            exists[name] = (file_exists := (f := getattr(self.files, name)).is_file())
            if file_exists and f.stat().st_size == 0:
                term.log.warning(f"{f} appears to be empty")

        if final := self.files.final(self.opts):
            exists["final"] = (file_exists := final.is_file())
            # size of 'barcode count'
            if file_exists and final.stat().st_size <= 14:
                print(final)
                print(final.stat().st_size)
                term.log.warning(f"{final} appears to be empty")
        else:
            exists["final"] = False
        self.files_exist = exists
        return exists

    @status_check
    def _filter(self) -> bool | None:
        json, html = (
            self.opts.pipeline / "qc" / f"{self.name}.{ext}" for ext in ("json", "html")
        )
        msg = "quality filtering reads with fastp"
        (self.opts.pipeline / "qc").mkdir(exist_ok=True)

        if not check_output(self.files.quality, msg):
            command = (
                "fastp "
                f"-i {self.fastq} "
                f"-o {self.files.quality} "
                f"-q {self.opts.quality} "
                f"-u {self.opts.unqualified_percent} "
                f"-w {self.opts.threads} "
                f"-h {html} "
                f"-j {json} "
                f"{self.opts.fastp_args or ''} "
            )
            with term.process(msg):
                return run_cmd(
                    command, self.name, self.files.quality, self.opts.verbose
                )

    @status_check
    def _cutadapt(
        self,
    ) -> bool | None:
        """perform quality filtering and extraction"""

        msg = "extracting barcodes with cutadapt"
        adapter_string = (
            f"-g {self.opts.upstream_adapter} -a {self.opts.downstream_adapter}"
            if self.opts.unlinked_adapters
            else f"-g {self.opts.upstream_adapter}...{self.opts.downstream_adapter}"
        )

        if self.opts.skip_trimming:
            shutil.copy(self.files.quality, self.files.barcode_fastq)

        if not check_output(
            self.files.quality.with_suffix(".barcode.fastq"),
            msg,
        ):
            command = (
                "cutadapt "
                f"-e {self.opts.error} "
                f"-j {self.opts.threads} "
                f"--minimum-length={self.opts.length - self.opts.distance} "
                f"--maximum-length={self.opts.length + self.opts.distance} "
                f"{adapter_string} "
                f"{self.opts.cutadapt_args or ''} "
                f"-o {self.files.barcode_fastq} {self.files.quality}"
            )
            with term.process(msg):
                return run_cmd(
                    command,
                    self.name,
                    self.files.barcode_fastq,
                    self.opts.verbose,
                )

    @status_check
    def _fast2tsv(self) -> bool | None:
        if not check_output(self.files.barcodes, "converting fastq to tsv"):
            return fastq_to_tsv(self.files.barcode_fastq, self.files.barcodes)

    @status_check
    def _starcode(self) -> bool | None:
        """cluster the barcodes using starcode"""

        msg = "clustering barcodes with starcode"

        if not check_output(self.files.clustered, msg):
            command = (
                f"starcode -d {self.opts.distance} -r {self.opts.ratio} "
                f"-t {self.opts.threads} -i {self.files.barcode_fastq} -o {self.files.clustered}"
            )
            with term.process(msg):
                return run_cmd(
                    command, self.name, self.files.clustered, self.opts.verbose
                )

    def _read_filter(self) -> None:
        if read_filter(self.name, self.opts):
            self.status = SampleStatus.WARN
        else:
            self.status = SampleStatus.COMPLETE


class MergeSample(Sample):
    def __init__(self, fastqR1: Path, fastqR2: Path, opts: PycashierOpts) -> None:
        name = fastqR1.name.split(".")[0]
        self.fastqR1 = fastqR1
        self.fastqR2 = fastqR2
        self.name = fastqR1.name.split(".")[0]
        self.merged = opts.output / f"{self.name}.merged.raw.fastq"
        self.steps = (self._fastp_merge,)
        super().__init__(name, opts)

    def check(self) -> Dict[str, bool]:
        exists = self.merged.is_file()
        if exists and self.merged.stat().st_size == 0:
            term.log.warning(f"{self.merged} appears to be empty")
        return {"final": exists}

    @status_check
    def _fastp_merge(
        self,
    ) -> bool | None:
        """merge a single sample with fastp"""

        msg = "merging paired end reads with fastp"
        if not check_output(self.merged, msg):
            command = (
                "fastp "
                f"-i {self.fastqR1}  "
                f"-I {self.fastqR2}  "
                f"-w {self.opts.threads} "
                f"-j {self.opts.pipeline}/merge_qc/{self.name}.json "
                f"-h {self.opts.pipeline}/merge_qc/{self.name}.html "
                f"--merged_out {self.merged} "
                f"{self.opts.fastp_args or ''}"
            )

            with term.process(msg):
                return run_cmd(command, self.name, self.merged, self.opts.verbose)


class ScrnaSample(Sample):
    def __init__(self, sam: Path, opts: PycashierOpts) -> None:
        name = sam.name.split(".")[0]

        self.sam = sam
        self.fastq = opts.pipeline / f"{name}.umi_cell_labeled.fastq"
        self.barcode_fastq = self.fastq.with_suffix(".barcode.fastq")
        self.barcodes = opts.output / f"{name}.umi_cell_labeled.barcode.tsv"
        self.steps = (self._sam_to_fastq, self._pysam_cutadapt, self._fast_to_tsv)
        super().__init__(name, opts)

    def check(self) -> Dict[str, bool]:
        return {
            f: getattr(self, f).is_file()
            for f in ("fastq", "barcode_fastq", "barcodes")
        }

    @status_check
    def _sam_to_fastq(self) -> bool | None:
        if not check_output(self.fastq, "converting sam to labeled fastq"):
            return sam_to_name_labeled_fastq(self.name, self.sam, self.fastq)

    @status_check
    def _pysam_cutadapt(
        self,
    ) -> bool | None:
        """extract barcodes from single cell data

        Args:
           status: Rich.console status to suspend for stderr printing.
        """
        msg = "extracting barcodes with cutadapt"
        adapter_string = (
            f"-g {self.opts.upstream_adapter} -a {self.opts.downstream_adapter}"
        )

        if not check_output(self.barcode_fastq, msg):
            command = (
                "cutadapt "
                f"-e {self.opts.error} "
                f"-j {self.opts.threads} "
                f"--minimum-length={self.opts.minimum_length} "
                f"--maximum-length={self.opts.length} "
                f"{adapter_string} "
                f"{self.opts.cutadapt_args or ''} "
                f"-o {self.barcode_fastq} {self.fastq}"
            )
            with term.process(msg):
                return run_cmd(
                    command, self.name, self.barcode_fastq, self.opts.verbose
                )

    @status_check
    def _fast_to_tsv(self) -> bool | None:
        if not check_output(self.barcodes, "converting labeled fastq to tsv"):
            return labeled_fastq_to_tsv(self.barcode_fastq, self.barcodes)



from __future__ import annotations

from pathlib import Path

import polars as pl

from .options import PycashierOpts
from .term import term
from .utils import get_filter_count


def filter_by_percent(
    file_in: Path, filter_percent: float, length: int, offset: int, outdir: Path
) -> bool | None:
    """filter clustered barcodes with nominal abundance cutoff

    Args:
        file_in: Path to clustered barcode counts.
        filter_percent: Minimum percent of total cutoff.
        length: Expected lenth of barcode.
        offset: Acceptable insertion or deletion from length in final sequences.
        output: Directory for final tsv files.
    """

    return filter_by_count(
        file_in, get_filter_count(file_in, filter_percent), length, offset, outdir
    )


def filter_by_count(
    file_in: Path, filter_count: int, length: int, offset: int, output: Path
) -> bool | None:
    """filter clusted barcodes with nominal abundance cutoff

    Args:
        file_in: Path to clustered barcode counts.
        filter_count: Minimum count cutoff.
        length: Expected lenth of barcode.
        offset: Acceptable insertion or deletion from length in final sequences.
        output: Directory for final tsv files.
    """

    final = output / f"{file_in.stem}.min{filter_count}_off{offset}{file_in.suffix}"

    # TODO: TRY/EXCEPT
    df = (
        pl.scan_csv(
            file_in, separator="\t", has_header=False, new_columns=["barcode", "count"]
        )
        .filter(
            (pl.col("count") > filter_count)
            & (
                (pl.col("barcode").str.len_chars().cast(pl.Int64) - length).abs()
                <= offset
            )
        )
        .collect()
    )

    df.write_csv(final, separator="\t")
    if df.height == 0:
        term.log.warning("no barcodes passed final length and abundance filters")
        return True


def read_filter(
    sample: str,
    opts: PycashierOpts,
) -> bool | None:
    """filter clusted barcodes with final abundance cutoff

    Args:
        sample: Name of the sample.
        opts: pycashier options
    """

    file_in = (
        opts.pipeline
        / f"{sample}.q{opts.quality}.barcodes.r{opts.ratio}d{opts.distance}.tsv"
    )
    if opts.filter_count is not None:
        term.log.debug(
            f"post-clustering filtering with [b]{opts.filter_count}[/] read cutoff"
        )
        return filter_by_count(
            file_in, int(opts.filter_count), opts.length, opts.offset, opts.output
        )

    else:
        term.log.debug(
            f"post-clustering filtering with [b]{opts.filter_percent}[/] % cutoff"
        )

        return filter_by_percent(
            file_in,
            opts.filter_percent,
            opts.length,
            opts.offset,
            opts.output,
        )



from __future__ import annotations

import multiprocessing
from typing import Dict, List

import click
from click.core import ParameterSource
from rich import box
from rich.table import Table
from rich.tree import Tree

from .options import PycashierOpts
from .sample import ExtractSample, MergeSample, Sample, ScrnaSample
from .term import term

SYS_THREADS = multiprocessing.cpu_count()


def show_sample_queue(
    samples: List[ExtractSample],
    completed_samples: List[ExtractSample],
    queue_all: bool,
    opts: PycashierOpts,
) -> None:
    """print formatted table for sample queue

    Args:
        queue_all: If true, skip output check and queue all samples.
        opts
    """

    table = Table(box=box.SIMPLE, header_style="bold cyan", collapse_padding=True)
    table.add_column(
        "sample",
        justify="center",
        style="green",
        no_wrap=True,
    )
    table.add_column(f"q{opts.quality}", justify="center")
    table.add_column("barcodes", justify="center")
    table.add_column(
        f"r{opts.ratio}d{opts.distance}",
        justify="center",
    )

    filter_expr = ("min(%)") if opts.filter_percent else "min(N)"
    table.add_column(f"{filter_expr}_off{opts.offset}", justify="center")

    if queue_all:
        for sample in samples:
            table.add_row(*([f"[bold yellow]{sample.name}"] + ["[yellow]Queued"] * 4))
    else:
        for sample in samples:
            color = "green" if sample.completed else "yellow"
            row = [f"[bold {color}]{sample.name}"] + [
                "[bold green]\u2713" if v else "[yellow]Queued"
                for v in sample.files_exist.values()
            ]
            table.add_row(*row)

    term.print(
        table,
        f"There are [bold cyan]{len(samples)-len(completed_samples)}[/] samples to finish processing.\n",
    )


def confirm_extract_samples(samples: List[ExtractSample], opts: PycashierOpts) -> None:
    queue_all = not opts.pipeline.is_dir()
    completed_samples = [sample for sample in samples if sample.completed]
    show_sample_queue(
        samples=samples,
        completed_samples=completed_samples,
        queue_all=queue_all,
        opts=opts,
    )

    if not opts.yes and (
        len(completed_samples) != len(samples)
        and not term.confirm("Continue with these samples?")
    ):
        term.quit(0)


def fmt_sample_list(samples: set[Sample]) -> str:
    return ", ".join(sorted((sample.name for sample in samples)))


def confirm_samples(
    samples: List[MergeSample] | List[ScrnaSample], opts: PycashierOpts
) -> None:
    completed = set(filter(lambda sample: sample.completed, samples))
    incomplete = set(samples) - completed

    term.print()
    if completed:
        term.print(f"[hl]Processed[/]: {fmt_sample_list(completed)}")
    if incomplete:
        term.print(f"[hl]To be processed[/]: {fmt_sample_list(incomplete)}")
    if not opts.yes and incomplete and not term.confirm("Continue with these samples?"):
        term.quit(0)


def print_params(ctx: click.Context) -> None:
    """show all parameters prior to run"""

    params = ctx.params
    verbose = params.pop("verbose")
    params.pop("save_config", None)
    params.pop("yes", None)
    params = {
        (k if k != "input_" else "input"): v for k, v in params.items() if v is not None
    }

    term.log.info("[bold]parameters")

    param_maps: Dict[str, Dict] = {"cli": {}, "config": {}, "default": {}}
    for k, v in params.items():
        source = ctx.get_parameter_source(k)
        if source == ParameterSource.DEFAULT:
            param_maps["default"][k] = v
        elif source == ParameterSource.DEFAULT_MAP:
            param_maps["config"][k] = v
        # elif source == ParameterSource.COMMANDLINE
        # params[input] is None and we don't use other ParameterSource's
        else:
            param_maps["cli"][k] = v

    for header, param_map in param_maps.items():
        if len(param_map) == 0:
            continue
        tree = Tree(f"[italic]{header}", style="bold")
        for k, v in param_map.items():
            tree.add(f"[bold][cyan]{k.replace('_', ' ')}[/cyan]: [yellow]{v}[/yellow]")

        with term._console.capture() as capture:
            term.print(tree)

        if header == "default" and not verbose:
            term.log.debug(capture.get())
        else:
            term.log.info(capture.get())

    if not verbose:
        term.print("[dim]use -v/--verbose to see all parameters")

    if not ctx.info_name == "receipt":
        if params["threads"] == 1 and params["threads"] <= SYS_THREADS / 4:
            term.print(
                f"[dim]Only using {params['threads']} of {SYS_THREADS} available threads..."
            )



#!/usr/bin/env python3

import os
import shlex
import subprocess
from pathlib import Path
from textwrap import indent

ROOT = Path(__file__).parent.parent


def run(cmd, **kwargs):
    return subprocess.run(shlex.split(cmd), **kwargs)


def get_output(cmd, **kwargs):
    return subprocess.check_output(shlex.split(cmd), text=True, **kwargs)


COMMAND_RST_TMPL = """
pycashier {command}
----------------------------------

.. tabs::

  .. tab:: image

    .. image:: svgs/{image_name}

  .. tab:: plaintext

    .. code-block::

{code}
"""


# TODO: add yartsu to docs dependencies?
def make_svg(command, image_path):
    run(f"viv run yartsu -- -o {image_path} -- pycashier {command} -h")


# make the pycashier plaintext not limited to 80 columns
env = os.environ.copy()
env["COLUMNS"] = "110"


def generate_rst(command=""):
    image_path = (
        ROOT
        / "docs/svgs"
        / (f"pycashier-{command}.svg" if command else "pycashier.svg")
    )
    make_svg(command, image_path)
    help_output = get_output(f"pycashier {command} -h", env=env)
    return COMMAND_RST_TMPL.format(
        command=command, image_name=image_path.name, code=indent(help_output, "      ")
    )


def main():
    cli_docs = "\n".join(
        ["=============", "CLI Reference", "=============", generate_rst()]
        + [generate_rst(cmd) for cmd in ["extract", "merge", "receipt", "scrna"]]
    )
    (ROOT / "docs/cli.rst").write_text(cli_docs)


if __name__ == "__main__":
    main()



from pathlib import Path

import click
import tomlkit

from ._checks import check_file_permissions
from .term import term


def save_params(ctx: click.Context) -> None:
    """save parameters to config file

    Args:
        ctx: Click context.
    """
    cmd = ctx.info_name
    all_params = {k: v for k, v in ctx.params.items() if v}
    save_type = all_params.pop("save_config")
    try:
        config_file = Path(ctx.obj["configfile"])
    except TypeError:
        raise click.BadParameter("use `--save-config` with a specified `--config`")

    if config_file.is_file():
        term.print(f"Updating current config file at [hl]{config_file}")
        try:
            with config_file.open("r") as f:
                config = tomlkit.load(f)
        except Exception as e:
            term.print(
                f"[ConfigFileError] failed to load config file: {config_file}", err=True
            )
            term.print(e, err=True)
            term.quit()

    else:
        term.print(f"Staring a config file at [hl]{config_file}")
        config = tomlkit.document()

    if save_type == "explicit":
        params = {}
        for k, v in all_params.items():
            print(ctx.get_parameter_source(k))
            if param_source := ctx.get_parameter_source(k):
                if param_source.value != 3:
                    params[k] = v
    else:
        params = all_params.copy()

    # use readable name for input
    if "input_" in params:
        params["input"] = params.pop("input_")

    # sanitize the path's for writing to toml
    for k in ["input", "pipeline", "output"]:
        if k in params.keys():
            params[k] = str(params[k])

    config[cmd] = params  # type: ignore

    null_hints = {"extract": ["filter_count", "fastp_args"], "merge": ["fastp_args"]}
    if cmd in null_hints.keys() and save_type == "full":
        for param in null_hints[cmd]:
            if param not in params.keys():
                config[cmd].add(tomlkit.comment(f"{param} ="))  # type: ignore

    with config_file.open("w") as f:
        f.write(tomlkit.dumps(config))

    term.print("Exiting...")
    ctx.exit()


def load_params(ctx: click.Context, param: str, filename: Path) -> None:
    """load parameters from config file

    Args:
        ctx: Click context.
        param: Invoked command and config table head.
        filename: Config filename.
    """
    ctx.obj = {"configfile": filename}

    if not filename or ctx.resilient_parsing:
        return

    ctx.default_map = {}

    check_file_permissions()
    config_file = Path(filename)
    if config_file.is_file():
        global_params = {}
        try:
            with config_file.open("r") as f:
                params = tomlkit.load(f)
        except Exception as e:
            term.print(
                f"[ConfigFileError] failed to load config file: {config_file}", err=True
            )
            term.print(e, err=True)
            term.quit()

        if params:
            global_params = params.pop("global", {})

            ctx.default_map = params.get(ctx.info_name, {})

            # use not shadowing name for input
            if ctx.default_map and "input" in ctx.default_map:
                ctx.default_map["input_"] = ctx.default_map.pop("input")

            # populate 'default_map' with global params
            for k, v in global_params.items():
                ctx.default_map.setdefault(k, v)  # type: ignore
        if global_params:
            ctx.obj.update(**{"global": global_params})
        ctx.obj["config-loaded"] = True

    # using default pycashier can cause unexpected behavior?
    elif config_file != Path("pycashier.toml"):
        term.print(
            f"[InputError]: Specified config file ({filename}) does not exist.",
            err=True,
        )
        term.quit()



from __future__ import annotations

from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
)

import click

from ._checks import pre_run_check
from .config import load_params
from .utils import validate_filter_args


def init_check(ctx: click.Context, param: str, check: bool) -> None:
    if not check:
        pre_run_check(command=ctx.to_info_dict()["command"]["name"])


def add_options(
    options: List[Callable],
) -> Callable[[click.decorators.FC], click.decorators.FC]:
    def _add_options(func: click.decorators.FC) -> click.decorators.FC:
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


class Option:
    """custom options class to wrap click.option"""

    def __init__(
        self,
        param_decls: Sequence[str],
        help: str = "",
        default: str | int | float | Path | None = None,
        show_default: bool = False,
        is_flag: bool = False,
        type: Any = None,
        required: bool = False,
        params: Dict[str, Any] = {},
        name: str = "",
        category: Optional[str] = None,
    ):
        self.param_decls = param_decls
        self.help = help
        self.default = default
        self.show_default = show_default
        self.is_flag = is_flag
        self.type = type
        self.required = required
        self.params = params
        self.category = category

        if name:
            self.name = name
        else:
            self.name = max(self.param_decls, key=len).lstrip("-")

    def get_click_option(self) -> Callable[[click.decorators.FC], click.decorators.FC]:
        return click.option(
            *self.param_decls,
            help=self.help,
            default=self.default,
            show_default=self.show_default,
            is_flag=self.is_flag,
            type=self.type,
            required=self.required,
            **self.params,
        )

    # def get_click_option_decorator(self, func: click.decorators.FC) ->click.decorators.FC:
    #     func = self.get_click_option()(func)
    #     return func

    def long(self) -> str:
        """return longest param declaration"""
        return max(self.param_decls, key=len)


general_opts = (
    "verbose",
    "config",
    "save-config",
    "skip-init-check",
    "log-file",
    "pipeline",
    "samples",
)


class OptionMap:
    def __init__(
        self, options: List[Option], subcmd_ref: Dict[str, Tuple[str, ...]]
    ) -> None:
        self.opts = {opt.name: opt for opt in options}
        self.subcmds = {k: [self.opts[opt] for opt in v] for k, v in subcmd_ref.items()}

    def long_by_category(self, category: str) -> Generator[str, None, None]:
        return (opt.long() for opt in self.opts.values() if opt.category == category)

    def get(self, param: str) -> Any:
        """return default value for param"""
        try:
            opt = self.opts[param]
        except KeyError:
            raise ValueError(f"{param} does not exist")
        return opt.default


_duplicate_options_args_kwargs = {
    "fastp-args": (
        ["-fa", "--fastp-args"],
        dict(
            help="additional arguments passed to fastp",
            show_default=True,
        ),
    ),
    "input": (
        ["-i", "--input", "input_"],
        dict(
            required=True,
            type=click.Path(exists=True, file_okay=False, path_type=Path),
            category="input/output",
        ),
    ),
    "output": (
        ["-o", "--output"],
        dict(
            help="output directory",
            show_default=True,
            type=click.Path(file_okay=False, path_type=Path),
            category="input/output",
        ),
    ),
}


def _make_deduplicated_opt(option: str, subcmd: str, **unique_kwargs: Any) -> Option:
    args, kwargs = _duplicate_options_args_kwargs[option]
    kwargs.update(unique_kwargs)
    kwargs.update({"name": f"{option}-{subcmd}"})
    return Option(args, **kwargs)  # type: ignore


_options = [
    Option(
        ["-o", "--output"],
        help="output directory",
        default="./outs",
        show_default=True,
        type=click.Path(file_okay=False, path_type=Path),
        category="input/output",
    ),
    Option(
        ["-p", "--pipeline"],
        help="pipeline directory",
        default="./pipeline",
        show_default=True,
        type=click.Path(file_okay=False, path_type=Path),
        category="input/output",
    ),
    Option(
        ["-s", "--samples"],
        help="comma separated list of samples to process",
        type=str,
        category="input/output",
    ),
    Option(
        ["-v", "--verbose"],
        help="show more output, set log level to debug",
        is_flag=True,
        category="general",
    ),
    Option(
        ["-t", "--threads"],
        help="number of cpu cores to use",
        default=1,
        show_default=True,
        category="general",
    ),
    Option(
        ["-c", "--config"],
        help="read parameter values from config file",
        show_default=True,
        default=Path("pycashier.toml"),
        type=click.Path(dir_okay=False),
        params=dict(
            expose_value=False,
            is_eager=True,
            callback=load_params,
        ),
        category="general",
    ),
    Option(
        ["--save-config"],
        help="save current params to file specified by `--config`",
        type=click.Choice(["explicit", "full"], case_sensitive=False),
        category="general",
    ),
    Option(
        ["--skip-init-check"],
        help="skip runtime dependency check",
        is_flag=False,
        params=dict(
            hidden=True,
            expose_value=False,
            callback=init_check,
        ),
        category="hidden",
    ),
    Option(
        ["--log-file"],
        help=r"path to log file [dim]\[default: <pipeline-dir>/pycashier.log]",
        category="general",
        type=click.Path(dir_okay=False, path_type=Path),
    ),
    Option(
        ["-y", "--yes"],
        help="answer yes to prompts",
        is_flag=True,
        category="general",
    ),
    Option(
        ["-e", "--error"],
        help="error tolerance supplied to cutadapt",
        default=0.1,
        show_default=True,
        category="trim",
    ),
    Option(
        ["-l", "--length"],
        help="target length of extracted barcode",
        default=20,
        show_default=True,
        category="trim",
    ),
    Option(
        ["-ua", "--upstream-adapter"],
        help="5' sequence flanking barcode",
        default="GTGGAAAGGACGAAACACCG",
        category="trim",
    ),
    Option(
        ["-da", "--downstream-adapter"],
        help="3' sequence flanking barcode",
        default="GTTTTAGAGCTAGAAATAGC",
        category="trim",
    ),
    Option(
        ["--unlinked-adapters"],
        help="run cutadapt using unlinked adapters",
        is_flag=True,
        category="trim",
    ),
    Option(
        ["--skip-trimming"],
        help="skip cutadapt trimming entirely and use reads as-is",
        is_flag=True,
        category="trim",
    ),
    Option(
        ["-q", "--quality"],
        help="minimum PHRED quality for filtering reads",
        default=30,
        show_default=True,
        category="quality",
    ),
    Option(
        ["-up", "--unqualified-percent"],
        help="minimum percent of bases which can be below quality threshold",
        default=20,
        show_default=True,
        category="quality",
    ),
    Option(
        ["-r", "--ratio"],
        help="ratio to use for message passing clustering",
        default=3,
        show_default=True,
        category="cluster",
    ),
    Option(
        ["-d", "--distance"],
        help="levenshtein distance for clustering",
        default=1,
        show_default=True,
        type=click.IntRange(1, 8),
        category="cluster",
    ),
    Option(
        ["-fc", "--filter-count"],
        help="minium nominal number of reads",
        type=int,
        category="filter",
    ),
    Option(
        ["-fp", "--filter-percent"],
        help="minimum percentage of total reads",
        default=0.005,
        show_default=True,
        category="filter",
    ),
    Option(
        ["--offset"],
        help="length offset from target barcode length post-clustering",
        default=1,
        show_default=True,
        category="filter",
    ),
    Option(
        ["-ml", "--minimum-length"],
        help="minimum length of extracted barcode",
        default=10,
        show_default=True,
        category="scrna",
    ),
    Option(
        ["--no-overlap"],
        help="skip per lineage overlap column",
        is_flag=True,
    ),
    Option(
        ["-ca", "--cutadapt-args"],
        help="additional arguments passed to cutadapt",
        type=str,
        default="--max-n=0 -n 2 --trimmed-only",
        show_default=True,
        category="trim",
    ),
]


_options += [
    _make_deduplicated_opt(
        "fastp-args",
        "extract",
        default="--dont_eval_duplication",
        category="quality",
    ),
    _make_deduplicated_opt(
        "fastp-args",
        "merge",
        default="-m -c -G -Q -L",
        category="merge",
    ),
    _make_deduplicated_opt(
        "input",
        "merge",
        help="source directory containing gzipped R1 and R2 fastq files",
    ),
    _make_deduplicated_opt(
        "input",
        "extract",
        help="source directory containing fastq files",
    ),
    _make_deduplicated_opt(
        "input",
        "scrna",
        help="source directory containing sam files from scRNA-seq",
    ),
    _make_deduplicated_opt(
        "input",
        "receipt",
        default="./outs",
        show_default=True,
    ),
    _make_deduplicated_opt(
        "output",
        "merge",
        default="./mergedfastqs",
    ),
    _make_deduplicated_opt(
        "output",
        "receipt",
        help="combined tsv of all samples found in input directory",
        type=click.Path(file_okay=True, dir_okay=False, path_type=Path),
        default="./combined.tsv",
    ),
]
optmap = OptionMap(
    _options,
    {
        "receipt": (
            "input-receipt",
            "output-receipt",
            "no-overlap",
            *general_opts,
        ),
        "extract": (
            "input-extract",
            "output",
            "quality",
            "unqualified-percent",
            "fastp-args-extract",
            "cutadapt-args",
            "error",
            "length",
            "upstream-adapter",
            "downstream-adapter",
            "unlinked-adapters",
            "skip-trimming",
            "ratio",
            "distance",
            "filter-count",
            "filter-percent",
            "offset",
            "threads",
            "yes",
            *general_opts,
        ),
        "merge": (
            "input-merge",
            "output-merge",
            "fastp-args-merge",
            "threads",
            "yes",
            *general_opts,
        ),
        "scrna": (
            "input-scrna",
            "output",
            "error",
            "length",
            "upstream-adapter",
            "downstream-adapter",
            "cutadapt-args",
            "minimum-length",
            "threads",
            "yes",
            *general_opts,
        ),
    },
)


class PycashierOpts:
    input_: Path
    threads: int
    pipeline: Path
    output: Path
    verbose: bool
    log_file: Path
    samples: Optional[str] = None
    fastp_args: str

    quality = optmap.get("quality")
    unqualified_percent = optmap.get("unqualified-percent")
    error = optmap.get("error")
    length = optmap.get("length")
    distance = optmap.get("distance")
    ratio = optmap.get("ratio")
    upstream_adapter = optmap.get("upstream-adapter")
    downstream_adapter = optmap.get("downstream-adapter")
    minimum_length = optmap.get("minimum-length")
    unlinked_adapters = optmap.get("unlinked-adapters")
    skip_trimming = optmap.get("skip-trimming")
    offset = optmap.get("offset")
    cutadapt_args = optmap.get("cutadapt-args")
    filter_count = optmap.get("filter-count")
    filter_percent = optmap.get("filter-percent")
    no_overlap = optmap.get("no-overlap")
    yes = optmap.get("yes")

    def __init__(self, **kwargs: Any) -> None:
        self.__dict__.update(kwargs)
        if self.log_file is None:
            self.log_file = self.pipeline / "pycashier.log"

    def update_filter(self, ctx: click.Context) -> None:
        filter = validate_filter_args(ctx)
        self.filter_percent = filter.get("filter_percent")
        self.filter_count = filter.get("filter_count")
